import asyncio
import fractions
import logging

import av
import numpy as np

from outspeed.plugins.base_plugin import Plugin
from outspeed.streams import AudioStream, ByteStream

logger = logging.getLogger(__name__)


class AudioConverter(Plugin):
    def __init__(self, sample_rate=16000, channel_layout="mono", format="s16"):
        self.input_sample_rate = sample_rate
        self.input_channel_layout = channel_layout
        self.input_format = format
        self.output_queue = AudioStream()
        self.audio_samples = 0
        self.output_audio_sample_rate = 48000
        # audio_samples = 0
        self.output_audio_time_base = fractions.Fraction(1, self.output_audio_sample_rate)
        self.output_audio_resampler = av.AudioResampler(
            format="s16",
            layout="stereo",
            rate=self.output_audio_sample_rate,
            frame_size=int(self.output_audio_sample_rate * 0.020),
        )

    async def run(self, input_queue: ByteStream) -> AudioStream:
        self.input_queue = input_queue
        self._task = asyncio.create_task(self.convert_bytes_to_frame())
        return self.output_queue

    async def convert_bytes_to_frame(self):
        audio_data = b""
        while True:
            chunk = await self.input_queue.get()
            audio_data += chunk
            if len(audio_data) < 2:
                continue
            if len(audio_data) % 2 != 0:
                array = np.frombuffer(audio_data[:-1], dtype=np.int16).reshape(1, -1)  # mono has 1 channel
                audio_data = audio_data[-1:]
            else:
                array = np.frombuffer(audio_data, dtype=np.int16).reshape(1, -1)  # mono has 1 channel
                audio_data = b""

            # Create a new AudioFrame from the NumPy array
            frame = av.AudioFrame.from_ndarray(array, format=self.input_format, layout=self.input_channel_layout)
            frame.sample_rate = self.input_sample_rate

            for nframe in self.output_audio_resampler.resample(frame):
                # fix timestamps
                nframe.pts = self.audio_samples
                nframe.time_base = self.output_audio_time_base
                self.audio_samples += nframe.samples
                self.output_queue.put_nowait(nframe)



import asyncio
import logging
from typing import List, Union

from outspeed.streams import AudioStream, ByteStream, Stream, TextStream, VideoStream


def merge(input_queues: List[Stream]) -> Union[AudioStream, VideoStream, TextStream, ByteStream]:
    """
    Merge multiple input streams of the same type into a single output stream.

    This function takes a list of input streams and combines them into a single output stream
    of the same type. It supports AudioStream, VideoStream, TextStream, and ByteStream types.

    Args:
        input_queues (List[Stream]): A list of input streams to be merged.

    Returns:
        Union[AudioStream, VideoStream, TextStream, ByteStream]: A single output stream
        containing data from all input streams.

    Raises:
        ValueError: If the input queues are not all of the same type or if an unsupported
        stream type is provided.
    """
    # Initialize the output queue
    output_queue: Union[AudioStream, VideoStream, TextStream, ByteStream] = None

    # Determine the type of the output queue based on the input queue type
    if isinstance(input_queues[0], AudioStream):
        output_queue = AudioStream()
    elif isinstance(input_queues[0], VideoStream):
        output_queue = VideoStream()
    elif isinstance(input_queues[0], TextStream):
        output_queue = TextStream()
    elif isinstance(input_queues[0], ByteStream):
        output_queue = ByteStream()
    else:
        raise ValueError(f"Invalid input queue type: {type(input_queues[0])}")

    async def get(queue: Stream) -> None:
        """
        Asynchronous function to continuously get items from an input queue
        and put them into the output queue.

        Args:
            queue (Stream): The input stream to read from.
        """
        try:
            while True:
                item = await queue.get()
                output_queue.put_nowait(item)
        except asyncio.CancelledError:
            pass
        except RuntimeError:
            # This is expected when event loop is closed
            pass
        except Exception as e:
            logging.error(f"Error in merge: {e}")

    # Create a task for each input queue to continuously read and merge data
    for q in input_queues:
        asyncio.create_task(get(q))

    return output_queue



import asyncio
import logging
import threading
import time

from outspeed.data import AudioData
from outspeed.plugins.base_plugin import Plugin
from outspeed.plugins.VAD.silero_model import SileroVADModel
from outspeed.streams import AudioStream, VADStream
from outspeed.utils.audio import calculate_audio_volume, exp_smoothing
from outspeed.utils.vad import VADState


class SileroVAD(Plugin):
    def __init__(
        self,
        min_speech_duration_seconds: float = 0.2,
        min_silence_duration_seconds: float = 0.25,
        activation_threshold: float = 0.5,
        min_volume: float = 0.6,
    ):
        self.model = SileroVADModel(sample_rate=8000, num_channels=1)

        self._activation_threshold = activation_threshold

        self._loop: asyncio.AbstractEventLoop = asyncio.get_running_loop()
        self.output_queue = VADStream()
        self.user_speaking = False

        self._vad_buffer = b""
        self._num_channels = 1
        self._sample_rate = 8000

        # Volume exponential smoothing
        self._smoothing_factor = 0.2
        self._prev_volume = 0
        self._min_volume = min_volume

        self._silero_chunk_size = 512 if self._sample_rate == 16000 else 256
        self._silero_chunk_bytes = self._silero_chunk_size * self._num_channels * 2

        self._min_speech_duration_seconds = min_speech_duration_seconds
        self._min_silence_duration_seconds = min_silence_duration_seconds
        self._speech_duration_seconds = 0
        self._silence_duration_seconds = 0
        self._vad_state = VADState.QUIET
        self._prev_vad_state = VADState.QUIET

    def run(self, input_queue: AudioStream) -> VADStream:
        self.input_queue = input_queue
        self._vad_thread = threading.Thread(target=self.execute_vad, daemon=True)
        self._vad_thread.start()
        logging.debug("Starting SileroVAD execution")
        return self.output_queue

    def _get_smoothed_volume(self, audio: bytes) -> float:
        volume = calculate_audio_volume(audio, self._sample_rate)
        return exp_smoothing(volume, self._prev_volume, self._smoothing_factor)

    def execute_vad(self):
        try:
            asyncio.run_coroutine_threadsafe(self.output_queue.put(self._vad_state), self._loop)
            while True:
                future = asyncio.run_coroutine_threadsafe(self.input_queue.get(), self._loop)
                audio_data: AudioData = future.result()
                if audio_data is None or not isinstance(audio_data, AudioData):
                    continue
                buffer = audio_data.resample(self._sample_rate, self._num_channels).get_bytes()
                self._vad_buffer += buffer

                if len(self._vad_buffer) < self._silero_chunk_bytes:
                    continue

                while len(self._vad_buffer) >= self._silero_chunk_bytes:
                    audio_frames = self._vad_buffer[: self._silero_chunk_bytes]
                    self._vad_buffer = self._vad_buffer[self._silero_chunk_bytes :]

                    confidence = self.model.voice_confidence(audio_frames)

                    volume = self._get_smoothed_volume(audio_frames)
                    self._prev_volume = volume

                    speaking = confidence >= self._activation_threshold and volume >= self._min_volume

                    duration_seconds = self._get_speech_duration_seconds(audio_frames)
                    if speaking:
                        if self._vad_state == VADState.QUIET or self._vad_state == VADState.STOPPING:
                            self._vad_state = VADState.STARTING
                        self._speech_duration_seconds += duration_seconds
                    else:
                        if self._vad_state == VADState.STARTING or self._vad_state == VADState.SPEAKING:
                            self._vad_state = VADState.STOPPING
                        self._silence_duration_seconds += duration_seconds

                    if (
                        self._vad_state == VADState.STARTING
                        and self._speech_duration_seconds >= self._min_speech_duration_seconds
                    ):
                        self._vad_state = VADState.SPEAKING
                        self._speech_duration_seconds = 0
                        self._silence_duration_seconds = 0
                        logging.debug("Speech detected, transitioning to SPEAKING state")

                    elif (
                        self._vad_state == VADState.STOPPING
                        and self._silence_duration_seconds >= self._min_silence_duration_seconds
                    ):
                        self._vad_state = VADState.QUIET
                        self._silence_duration_seconds = 0
                        self._speech_duration_seconds = 0
                        logging.debug("Silence detected, transitioning to QUIET state")

                    if self._vad_state != self._prev_vad_state:
                        logging.debug(f"VAD state: {self._vad_state}")
                        self._prev_vad_state = self._vad_state
                        asyncio.run_coroutine_threadsafe(self.output_queue.put(self._vad_state), self._loop)
        except Exception as e:
            logging.error(f"Error in VAD execution: {str(e)}")

    def _get_speech_duration_seconds(self, audio_frames: bytes) -> float:
        return len(audio_frames) / (self._sample_rate * self._num_channels * 2)
